#!/usr/bin/env node

const chalk = require('chalk');
const ora = require('ora');
const program = require('commander');
const inquirer = require('inquirer');
const spawn = require('child_process').spawn;
const shell = require('shelljs');
var path = require("path");
const nconf = require('nconf');
const semver = require('semver');
const spinner = ora('Installing ');
const os = require('os');
const fs = require('fs');

var remoteSyncURL = "http://biserok.com.ua/api/updates"; ////TODO: add real server path!
var versionCheckURL = "http://biserok.com.ua/api/v1/version.json"; //TODO: add real server path!
var workDir = path.resolve('.');

var defaultThemePath = 'Public/Templates/default/default';
var defaultAdminThemePath = 'Public/Templates/default/admin';
var localizationsPath = 'Localizations';  

var tmp = shell.tempdir();

program
  .option('-d, --debug', 'builds with debug symbols.') // with production env?
  //.option('-b, --build', 'automatically build project after updating.') // 
  .option('-o, --only-dependencies', 'updates only swift dependendencies.') // update only dependencies
  .parse(process.argv);

nconf.argv()
  .env()
  .file({ file: workDir + '/Config/app.json' });

var regenerateXcodeQuestion = [{
  type: 'confirm',
  name: 'xcode',
  message: 'Would you like to regenerate your xcode project now?'
}];

var openXcodeQuestion = [{
  type: 'confirm',
  name: 'open',
  message: 'Open Xcode project?'
}];

//check dependencies
if (!shell.which('curl', { stdio: 'inherit' })) {
  shell.exec('sudo apt-get install curl');
}

if (!shell.which('rclone', { stdio: 'inherit' })) {
  console.log('Installing rclone...')
  shell.exec('curl https://rclone.org/install.sh | sudo bash', { stdio: 'inherit' });
}

var currentVersion = getCurrentVersion();
var currentCoreVersion = getCurrentCoreVersion();
var latestVersion = getLatestVersion(); 

//check version
checkVersion();


function getCurrentVersion() {
  return nconf.get('version');
}

function getLatestVersion() {
  var response = shell.exec(`curl -X GET ${versionCheckURL}`, {
    stdio: 'inherit',
    silent: true
  }).stdout

  var config = JSON.parse(response);
  if (!('version' in config)  || !('core' in config)) {
    spinner.fail('Could not check latest version');
    process.exit(1);
  }

  if (config['version'] !== undefined || config['core'] !== undefined) {
    //return config['version']
    return config
  } else {
    spinner.fail('Could not check latest version');
    process.exit(1);
  }
}

function checkVersion() {
  if (program.onlyDependencies) {
    update();
  } else {
    //var currentVersion = getCurrentVersion();
    //var latestVersion = getLatestVersion();
    if (semver.gte(currentVersion, latestVersion['version'])) {
      spinner.succeed('Already up-to-date.');
      process.exit(0);
    } else {
      update(latestVersion['version']);
    }
  }
}

function update(version) {
  console.log('Updating...')
  if(!program.onlyDependencies) {
    makeBackup();
    updateCoreVersion();    
  }    
  let child = spawn('swift', ['package', 'update'], {
    stdio: 'inherit',
    shell: true
  });
  child.on('exit', (code) => {
    if (code == 0) {
      if (!program.onlyDependencies) {
        build(version);
      } else {
        if (os.platform() == 'darwin') {
          regenerateXcodeProject();
        }
      }
    } else {
      if (!program.onlyDependencies) {
        rollback();
        process.exit(code);
      } else {
        spinner.fail('Update failed');        
        process.exit(code);
      }
    }
  });
}

function build(version) {
  var args = ['swift', 'build', '--enable-prefetching'];
  if (program.debug) {
    args.push('-Xswiftc');
    args.push('-g');
  } else {
    args.push('-c');
    args.push('release');
  }
  let child = spawn('npx', args, {
    stdio: 'inherit',
    shell: true
  });
  child.on('exit', (code) => {
    if (code == 0) {
      syncData(version);
    } else {
      rollback();
      process.exit(code);
    }
  });
}

function syncData(version) {    
  var syncDefaultTheme = `RCLONE_CONFIG_ZZ_TYPE=http RCLONE_CONFIG_ZZ_URL=${remoteSyncURL}/${version} rclone sync --exclude "/custom/" zz:${defaultThemePath} ${workDir}/${defaultThemePath} --delete-after --quiet`;  
  var syncDefaultAdminTheme = `RCLONE_CONFIG_ZZ_TYPE=http RCLONE_CONFIG_ZZ_URL=${remoteSyncURL}/${version} rclone sync --exclude "/custom/" zz:${defaultAdminThemePath} ${workDir}/${defaultAdminThemePath} --delete-after --quiet`;  
  var syncLocalizations = `RCLONE_CONFIG_ZZ_TYPE=http RCLONE_CONFIG_ZZ_URL=${remoteSyncURL}/${version} rclone sync --exclude "/custom/" zz:${localizationsPath} ${workDir}/${localizationsPath} --delete-after --quiet`;

  if (shell.exec(syncDefaultTheme, { stdio: 'inherit' }).code !== 0) {
    rollback();
    shell.exit(1);
  }
  if (shell.exec(syncDefaultAdminTheme, { stdio: 'inherit' }).code !== 0) {
    rollback();
    shell.exit(1);
  }
  if (shell.exec(syncLocalizations, { stdio: 'inherit' }).code !== 0) {
    rollback();
    shell.exit(1);
  }
  updateVersion(version);
}

function regenerateXcodeProject() {
  console.log(chalk.cyan('Changes to dependencies usually require Xcode to be regenerated.'));
  inquirer.prompt(regenerateXcodeQuestion).then(answers => {
    if (answers.xcode) {
      if (shell.exec('swift package generate-xcodeproj', {
        stdio: 'inherit'
      }).code !== 0) {
        spinner.fail('Could not generate Xcode project');
      } else {
        inquirer.prompt(openXcodeQuestion).then(answers => {
          if (answers.open) {
            console.log('Opening Xcode project...');
            shell.exec('open *.xcodeproj', {
              stdio: 'inherit'
            })
          } else {
            finish();
          }
        });
      }
    } else {
      finish();
    }
  });
}

function updateVersion(version) {  
  //var latestVersion = getLatestVersion();
  //nconf.set('version', latestVersion['version']);
  nconf.set('version', version);
  nconf.save(function (err) {
    fs.readFile(workDir + '/Config/app.json', function (err, data) {
      var json = JSON.parse(data);
      console.log('Updated to version ' + json['version']);
      if (os.platform() == 'darwin') {
        regenerateXcodeProject();
      } else {
        finish();
      }
    });
  });
}

function updateCoreVersion() {
  //var currentVersion = getCurrentCoreVersion();
  //var latestVersion = getLatestVersion();  
  console.log('Updating core version...');
 // console.log('Latest core version: ' + latestVersion['core']);  
  if (!semver.gte(currentCoreVersion, latestVersion['core'])) {
    //replace version
    var str = `.package(url: \"https://github.com/vapor/vapor.git\", .upToNextMajor(from: \"${currentCoreVersion}\")),`;
    var updatedStr = `.package(url: \"https://github.com/vapor/vapor.git\", .upToNextMajor(from: \"${latestVersion['core']}\")),`;
    var path = workDir + '/Package.swift';

    var updateVersion = `sed -i 's|${str}|${updatedStr}|g' ${path}`;    
    if (os.platform() == 'darwin') {
      updateVersion = `sed -i '' 's|${str}|${updatedStr}|g' ${path}`;
    }
    if (shell.exec(updateVersion, { stdio: 'inherit' }).code !== 0) {
      spinner.fail('Failed to update package.swift.');
      shell.exit(1);
    }
  }
}
/*
function rollbackCoreVersion(version) {
    //var latestVersion = getLatestVersion(); 
    console.log('Rollback core version');    
    //replace version
    var path = workDir + '/Package.swift';
    var file = fs.readFileSync(path, "utf8");
    var str = `.package(url: \"https://github.com/vapor/vapor.git\", .upToNextMajor(from: \"${latestVersion['core']}\")),`;
    var updatedStr = `.package(url: \"https://github.com/vapor/vapor.git\", .upToNextMajor(from: \"${currentCoreVersion}\")),`;    

    if (~file.indexOf(str)) {
      console.log('String was found in package.swift');
      var updateVersion = `sed -i 's|${str}|${updatedStr}|g' ${path}`;    
      if (os.platform() == 'darwin') {
        updateVersion = `sed -i '' 's|${str}|${updatedStr}|g' ${path}`;
      }
      if (shell.exec(updateVersion, { stdio: 'inherit' }).code !== 0) {
        spinner.fail('Failed to update package.swift.');
        shell.exit(1);
      }
    }
}
*/
function getCurrentCoreVersion() {
    var path = workDir + '/Package.swift';    
    var file = fs.readFileSync(path, "utf8");
    var start = '.package(url: \"https://github.com/vapor/vapor.git\", .upToNextMajor(from: \"'; //FIXME: add real path
    var end = '\")),';
    var version = file.split(start).pop().split(end).shift();
   // console.log('Core Version is ' + version);
    return version  

    /*
    fs.readFile(file, "utf8", function (err, data) {      
      var start = '.package(url: \"https://github.com/vapor/vapor.git\", .upToNextMajor(from: \"'; //FIXME: add real path
      var end = '\")),';
      var version = data.split(start).pop().split(end).shift();
      console.log('Core Version is ' + version);
      return version      
    });  
    */
}
//TODO: clean and build
function rollback() {
  //spinner.fail('Update failed');
  //console.log('Rollback...');
  //console.log('Previous version: ' + currentVersion);
  //console.log('Previous core version: ' + currentCoreVersion);

  restoreBackup();   
  spinner.fail('Update failed. Try to update your app later.');    
  process.exit(1);
  //var rollback = true;
//  rollbackCoreVersion(currentCoreVersion);  
  //rollBackToVersion(currentVersion)
//  update(currentVersion, true);
  //process.exit(1);
  //spinner.fail('Rollback failed. Try to update your app later');
  //process.exit(1);
  //откат изменений: возвращение старых файлов на место/синхронизация с предыдущей версией, откат на предыдущую версию CMS Core (exact) в package.swift, повторный билд, очистка кэша
}
/*
function rollbackToVersion(version) {

}
*/
//TODO: backup app.json?
function makeBackup() {
  var tmpDir = `${tmp}/Kami`;

  shell.mkdir('-p', `${tmpDir}/${defaultThemePath}`);
  shell.mkdir('-p', `${tmpDir}/${defaultAdminThemePath}`);
  
  //shell.mkdir('-p', `${tmpDir}/${localizationsPath}`);

  shell.cp('-Rf', `${workDir}/${defaultThemePath}`, `${tmpDir}/${defaultThemePath}`);
  shell.cp('-Rf', `${workDir}/${defaultAdminThemePath}`, `${tmpDir}/${defaultAdminThemePath}`);
  shell.cp('-Rf', `${workDir}/${localizationsPath}`, tmpDir);
  shell.cp('-Rf', `${workDir}/Package.swift`, tmpDir); 


  //shell.mkdir('-p', `${tmp}/Kami`);
  
  
  
  //logLevel = '--stats-log-level ERROR --log-level ERROR';
  //shell.exec(`rclone sync ${workDir}/${defaultThemePath} ${tmpDir}/${defaultThemePath} ${logLevel}`, { stdio: 'inherit' });
  //shell.exec(`rclone sync ${workDir}/${defaultAdminThemePath} ${tmpDir}/${defaultAdminThemePath} ${logLevel}`, { stdio: 'inherit' });
  //shell.exec(`rclone sync ${workDir}/${localizationsPath} ${tmpDir}/${localizationsPath} ${logLevel}`, { stdio: 'inherit' });
  //shell.exec(`rclone sync ${workDir}/Package.swift ${tmpDir} --quiet`, { stdio: 'inherit' });


  //if (shell.exec(syncDefaultTheme, { stdio: 'inherit' }).code !== 0) {
    //spinner.fail('Rollback failed');
    //shell.exit(1);
  //}


  //console.log('TEMP IS: ' + tmp);
  //console.log('TEMP DIRECTORY IS: ' + tmpDir);
  //console.log('DEFAULT THEME PATH IS: ' + `${workDir}/${defaultThemePath}`);
  //console.log('DEFAULT THEME TMP PATH IS: ' + `${tmpDir}/${defaultAdminThemePath}`);
  //shell.cp('-Rf', `${workDir}/${defaultThemePath}`, tmpDir);
  //shell.cp('-Rf', `${workDir}/${defaultAdminThemePath}`, tmpDir);
  //shell.cp('-Rf', `${workDir}/${localizationsPath}`, tmpDir);
  //shell.cp('-Rf', `${workDir}/Package.swift`, tmpDir);  
}

function restoreBackup() {
  var tmpDir = `${tmp}/Kami`;
  //shell.exec(`rclone sync ${tmpDir}/${defaultThemePath} ${workDir}/${defaultThemePath} --delete-after --quiet`, { stdio: 'inherit' });
  //shell.exec(`rclone sync ${tmpDir}/${defaultAdminThemePath} ${workDir}/${defaultAdminThemePath} --delete-after --quiet`, { stdio: 'inherit' });
  //shell.exec(`rclone sync ${tmpDir}/${localizationsPath} ${workDir}/${localizationsPath} --delete-after --quiet`, { stdio: 'inherit' });
  //shell.exec(`rclone sync ${tmpDir}/Package.swift ${workDir} --quiet`, { stdio: 'inherit' });

  shell.cp('-Rf', `${tmpDir}/${defaultThemePath}`, `${workDir}/${defaultThemePath}`);
  shell.cp('-Rf', `${tmp}/Kami/${defaultAdminThemePath}`, `${workDir}/${defaultAdminThemePath}`);
  shell.cp('-Rf', `${tmp}/Kami/${localizationsPath}`, `${workDir}`);
  shell.cp('-Rf', `${tmp}/Kami/Package.swift`, `${workDir}`);
  removeBackup();
}

function removeBackup() {
  shell.rm('-rf', `${tmp}/Kami`); 
}

function finish() {
  removeBackup();
  if (!program.onlyDependencies) {
    //updateVersion();
    spinner.succeed('Kami updated successfully. Restart your application now.');
    process.exit(0);
  } else {
    spinner.succeed('Done.');
    process.exit(0);
  }
}